--[[
	Epic documentation for me and whoever wants to create a companion script for this.
	player.willParaglide:
		When set to true, logic will look if the player is moving downwards. If so,
		this variable will be set to false, while player.paragliding will be set to true. 
	player.paragliding:
		When set to true, the paragliding logic is enabled for this player.
		Five things will set this to false and stop the logic:
			- Touching the ground.
			- Touching quicksand.
			- Death.
			- Getting hurt/spin-out.
			- The paraglider receiving damage.
	player.glideVerticalMovement:
		Controls-wise, it directs if the vertical movement should be upwards or
		downwards. Capped to PG_UPPERTICBOUNDS and PG_LOWERTICBOUNDS by logic.
	player.oldMoMz:
		Only used internally to halve the kart's vertical momentum by various
		degrees, then stored afterwards.
]]

-- IMPORTANT!! NEWER VERSION WILL OVERRIDE OLDER VERSIONS
local PARAGLIDERLOGICVERSION	= 2
if jglider == nil then
	rawset(_G, "jglider", {})
end

-- avoid redefiniton on updates
jglider.CVar_ParagliderDebug = $ or nil

if jglider.logicVer == nil or jglider.logicVer < PARAGLIDERLOGICVERSION then

if jglider.CVar_ParagliderDebug == nil then
	jglider.CVar_ParagliderDebug = CV_RegisterVar(
		{
			name = "paraglider_debug",
			defaultvalue = "Off",
			flags = CV_CHEAT,
			PossibleValue = CV_YesNo
		}
	)
end

jglider.isMobjValidPlayer = function(mobj)
	if mobj == nil 			then return false end
	if mobj.player == nil 	then return false end
	
	return true
end

local FRACUNIT 		= FRACUNIT
local TICRATE 		= TICRATE

local ANG1			= ANG1 
local ANGLE_90 		= ANGLE_90
local ANGLE_270 	= ANGLE_270

local BT_ACCELERATE = BT_ACCELERATE
local BT_BRAKE		= BT_BRAKE
local BT_ATTACK		= BT_ATTACK

local KITEM_SNEAKER			= KITEM_SNEAKER
local KITEM_ROCKETSNEAKER	= KITEM_ROCKETSNEAKER
local KITEM_INVINCIBILITY 	= KITEM_INVINCIBILITY

local WINDSECTOR	= 512
local FANSECTOR		= 20480

local STARTTIME 				= 6 * TICRATE + (3 * TICRATE / 4);
local ITEMTIME					= 8 * TICRATE

local PG_FALLRATE 				= 6 * FRACUNIT
local PG_MAXFALL				= 2 * FRACUNIT

local PG_PLAYERFALLINTENSITY	= 3 * FRACUNIT

local PG_UPPERTICBOUNDS			= FixedDiv(TICRATE, 2*FRACUNIT)
local PG_LOWERTICBOUNDS 		= -PG_UPPERTICBOUNDS

local PG_MAXSPEED				= 120 * FRACUNIT
local PG_MINSPEED				= 20 * FRACUNIT

local LETGOGLIDERTIMER			= TICRATE

local IMMEDIATEGLIDE			= ML_NOCLIMB

jglider.enableParagliding = function(line, playerMo, sector)
	if jglider.isMobjValidPlayer(playerMo) == false then return end

	if playerMo.willParaglide == true or
		playerMo.paragliding == true then return end

	if ((line.flags & IMMEDIATEGLIDE) == IMMEDIATEGLIDE) then
		playerMo.paragliding = true
	else
		playerMo.willParaglide = true
		S_StartSound(playerMo, sfx_ddash)
	end
end

jglider.removeParaglide = function(line, playerMo, sector)
	if jglider.isMobjValidPlayer(playerMo) == false then return end

	-- will not deploy
	if playerMo.willParaglide == true then
		playerMo.willParaglide = false
	end

	-- if already active...
	-- act as if player touched the ground
	-- also play a different sound
	if playerMo.paragliding == true then
		jglider.createDeadGlider(playerMo)
		jglider.resetGliderStuff(playerMo)
		S_StartSound(playerMo, sfx_s224)
	end
end

jglider.resetGliderStuff = function(playerMo)
	playerMo.paragliding 			= false
	playerMo.willParaglide 			= false
	playerMo.oldMoMz 				= 0
	playerMo.glideVerticalMovement 	= 0
	playerMo.letGoGliderTimer 		= 0

	if playerMo.paraglideMobj.valid then
		P_RemoveMobj(playerMo.paraglideMobj)
		playerMo.paraglideMobj = nil
	end
end

jglider.createDeadGlider = function(playerMo)
	local oldGlider = playerMo.paraglideMobj

	if oldGlider == nil or not oldGlider.valid then return end

	local deadgliderMobj = P_SpawnMobj(oldGlider.x, oldGlider.y, oldGlider.z,
		MT_JUG_DEADGLIDER)

	-- inherit various properties of the previous glider.
	deadgliderMobj.color	= playerMo.player.skincolor
	deadgliderMobj.angle 	= oldGlider.angle
	deadgliderMobj.scale	= oldGlider.scale
	
	-- Inherit the player's momentum
	deadgliderMobj.momx		= playerMo.momx / 2
	deadgliderMobj.momy		= playerMo.momy / 2
	deadgliderMobj.momz		= playerMo.oldMoMz

	-- lower fric
	deadgliderMobj.friction	= $ / 2
end

jglider.getTurnRateGlider = function(player)

	local fractionAng = FixedDiv(ANG1, FRACUNIT * 35)	
	local angleCorrectRate = ANG1 + (fractionAng * 5) -- 40 degrees per sec.
	-- speed 1 goes 39 degrees.
	-- speed 5 goes 35.
	-- speed 9 goes 30.
	local angleLoss	= fractionAng * player.kartweight
	angleCorrectRate = $ - angleLoss

	if player.kartstuff[k_driftboost] > 0 or
		player.kartstuff[k_sneakertimer] > 0 or
		player.kartstuff[k_invincibilitytimer] > 0 then

		angleCorrectRate = $ * 2
	end

	return angleCorrectRate
end

jglider.handleParaglide = function(playerMo)
	-- gets pressing up/down input (for moving the glide up and down)
	-- 1 is up, 0 is nothing, -1 is down
	local player = playerMo.player

	playerMo.glideVerticalMovement = playerMo.glideVerticalMovement or 0

	if leveltime > STARTTIME then
		local vertDir = player.kartstuff[k_throwdir] * P_MobjFlip(playerMo)

		-- handles the upwards movement input capping, capped to a certain degree
		if (vertDir > 0 and playerMo.glideVerticalMovement < PG_UPPERTICBOUNDS) or
			(vertDir < 0 and playerMo.glideVerticalMovement > PG_LOWERTICBOUNDS) then

				playerMo.glideVerticalMovement = $ + vertDir

			-- correcting a downwards movement by pressing up and viceversa is faster.
			if vertDir > 0 and playerMo.glideVerticalMovement < 0 then
				playerMo.glideVerticalMovement = $ + 1
			elseif vertDir < 0 and playerMo.glideVerticalMovement > 0 then
				playerMo.glideVerticalMovement = $ - 1
			end

		end
	end

	-- handling sideways movement	
	local playerAngle = playerMo.angle

	-- naturally correct this momentum angle to the player's
	local thrustAngle = R_PointToAngle2(0, 0, playerMo.momx, playerMo.momy)
	
	-- turn rate calc
	local angleCorrectRate = jglider.getTurnRateGlider(player)

	local angleMovementResult = angleCorrectRate
	local angleSidewaysMovementDiff = thrustAngle - playerAngle
	local angleAbsSidewaysMovement = abs(angleSidewaysMovementDiff)

	if angleAbsSidewaysMovement < angleCorrectRate then
		angleMovementResult = angleAbsSidewaysMovement
		-- print("angle less than one, correcting.")
	end

	if angleSidewaysMovementDiff > 0 then
		thrustAngle = $ - angleMovementResult
	elseif angleSidewaysMovementDiff < 0 then
		thrustAngle = $ + angleMovementResult
	end

	local thrustForce = R_PointToDist2(0, 0, playerMo.momx, playerMo.momy)

	-- bounds checking
	thrustForce = min(thrustForce, PG_MAXSPEED)

	-- let the player control the paraglider horizontally
	playerMo.momx = FixedMul(thrustForce, cos(thrustAngle))
	playerMo.momy = FixedMul(thrustForce, sin(thrustAngle))

	-- hold accel while in the air to accelerate in the air, 3/4s of the accel
	-- yup, makes a lot of sense.
	-- don't allow this at the start of levels to allow epic synchonized starts
	if leveltime > STARTTIME then
		-- and (player.cmd.buttons & BT_ACCELERATE) == BT_ACCELERATE 

		-- gotten straight from hardcode
		local accelMax		= 4000
		local ORIG_FRICTION	= 62914

		local accelValue 	= K_GetKartAccel(player)
		local speedValue	= K_GetKartSpeed(player, false)
		speedValue = FixedMul($, player.kartstuff[k_boostpower] + player.kartstuff[k_speedboost]);
		local oldSpeed		= R_PointToDist2(0, 0, player.rmomx, player.rmomy)

		-- also straight from hardcode
		local newSpeed = FixedDiv(FixedDiv(FixedMul(oldSpeed, accelMax - accelValue) +
			FixedMul(speedValue, accelValue), accelMax), ORIG_FRICTION)

		local finalSpeed = newSpeed - oldSpeed
		
		-- finalSpeed = $ / 8

		if finalSpeed < 0 then
			finalSpeed = 0
		end

		playerMo.momx = $ + P_ReturnThrustX(thrustAngle, finalSpeed)
		playerMo.momy = $ + P_ReturnThrustY(thrustAngle, finalSpeed)

		playerMo.momx = FixedMul(playerMo.momx, playerMo.friction)
		playerMo.momy = FixedMul(playerMo.momy, playerMo.friction)
	
	end

	-- print(playerMo.subsector.sector.special)
	local isPlayerInsideFanFoF = false
	local playerinFanFoF = P_ThingOnSpecial3DFloor(playerMo)

	if playerinFanFoF != nil then
		playerinFanFoF = ($.special & FANSECTOR) == FANSECTOR
		-- print("touching fan? " + isPlayerInsideFanFoF)
	end

	local playerSpecial = playerMo.subsector.sector.special
	local isPlayerInsideFanOrWindSector =
		playerSpecial & WINDSECTOR == WINDSECTOR or
		playerSpecial & FANSECTOR == FANSECTOR

	if not (isPlayerInsideFanOrWindSector or playerinFanFoF) then
		-- handling past momentum and fall formula
		playerMo.oldMoMz = playerMo.oldMoMz or 0

		local fallRateFormula = FixedMul(PG_FALLRATE +
			( FixedDiv(FRACUNIT * playerMo.glideVerticalMovement, PG_PLAYERFALLINTENSITY) ),
			mapobjectscale)

		-- prevent the kart from falling too fast
		local pg_maxfall_recalc = PG_MAXFALL

		if mapobjectscale != 1 then
			pg_maxfall_recalc = FixedMul(pg_maxfall_recalc, mapobjectscale)
		end

		if fallRateFormula < pg_maxfall_recalc then
			fallRateFormula = pg_maxfall_recalc
		end
		
		-- adjusting vertical momentum and set it
		-- thankfully the only thing that does not need to be mobjscaled... I hope.
		local momentum = FixedDiv((playerMo.momz - playerMo.oldMoMz), fallRateFormula)
		playerMo.momz = playerMo.oldMoMz + momentum
	end

	playerMo.oldMoMz = playerMo.momz
end

jglider.glidingLogic = function(playerMo)
	local player = playerMo.player

	if playerMo.willParaglide == true then


		-- funny visual indicator
		if leveltime % 2 then K_SpawnSparkleTrail(playerMo) end

		if playerMo.paraglideMobj == nil or
			playerMo.paraglideMobj.valid == false then				
			-- transformMomentumToLook(player)			
			local initialPos = FixedMul(16 * FRACUNIT, playerMo.scale) * P_MobjFlip(playerMo)

			local paraglideMobj 	= P_SpawnMobj(playerMo.x, playerMo.y, playerMo.z + initialPos,
				MT_JUG_PARAGLIDER)
			
			paraglideMobj.color		= player.skincolor
			paraglideMobj.target 	= playerMo
			paraglideMobj.angle 	= playerMo.angle

			paraglideMobj.scale		= 0
			paraglideMobj.destscale = playerMo.destscale
			paraglideMobj.scalespeed = playerMo.scalespeed

			playerMo.paraglideMobj 	= paraglideMobj
		end

		if playerMo.momz < 0 then
			playerMo.willParaglide = false		
			playerMo.paragliding = true

			playerMo.paraglideMobj.state = S_PAGL_SPAWN1
			S_StartSound(playerMo, sfx_paglgo)
		end
	end
	
	if not playerMo.paragliding 				then return end
	-- print(playerMo.paragliding == true)

	-- paraglide logic here.
	jglider.handleParaglide(playerMo)

	-- press item while holding a sneaker to enable the sneaker in mid-air
	-- could be extended easily to other items too...
	player.holdingItemBTN = $ or false
	if (player.cmd.buttons & BT_ATTACK) == BT_ATTACK and
		player.holdingItemBTN == false then
		
		if player.kartstuff[k_rocketsneakertimer] > 1 then
			K_DoSneaker(player, 2) -- taunt
			K_PlayBoostTaunt(playerMo)
			player.kartstuff[k_rocketsneakertimer] = $ - (2 * TICRATE)
			
			if player.kartstuff[k_rocketsneakertimer] < 1 then
				player.kartstuff[k_rocketsneakertimer] = 1
			end
		elseif player.kartstuff[k_itemtype] == KITEM_ROCKETSNEAKER then
			K_PlayBoostTaunt(playerMo)
			S_StartSound(playerMo, sfx_s3k3a)

			player.kartstuff[k_rocketsneakertimer] = ITEMTIME * 3
			player.kartstuff[k_itemamount] = $ - 1
			
			local prev = playerMo
			for i = 0,1 do
				local rocketSneakerMo = P_SpawnMobj(playerMo.x, playerMo.y, playerMo.z, MT_ROCKETSNEAKER)
				K_MatchGenericExtraFlags(rocketSneakerMo, playerMo)
				rocketSneakerMo.flags		= $ + MF_NOCLIPTHING
				rocketSneakerMo.angle 		= playerMo.angle
				rocketSneakerMo.threshold 	= 10 
				rocketSneakerMo.movecount 	= i % 2
				rocketSneakerMo.movedir		= i + 1
				rocketSneakerMo.lastlook 	= i + 1
				rocketSneakerMo.target 		= playerMo
				rocketSneakerMo.hprev 		= prev
				prev.hnext 					= rocketSneakerMo
				
				prev 						= rocketSneakerMo
			end

		elseif player.kartstuff[k_itemtype] == KITEM_SNEAKER then
			K_DoSneaker(player, 1) -- taunt
			K_PlayBoostTaunt(playerMo)
			player.kartstuff[k_itemamount] = $ - 1
		end

		-- this makes more sense to persist between deaths.
		if player.holdingItemBTN == false then
			player.holdingItemBTN = true
		end

	elseif (player.cmd.buttons & BT_ATTACK) ~= BT_ATTACK and
		player.holdingItemBTN == true then
		player.holdingItemBTN = false
	end

	-- hold brake for a sec to let go of the glider
	if (player.cmd.buttons & BT_BRAKE) == BT_BRAKE then
		playerMo.letGoGliderTimer = (playerMo.letGoGliderTimer or 0) + 1

		if playerMo.letGoGliderTimer >= LETGOGLIDERTIMER then
			jglider.createDeadGlider(playerMo)
			jglider.resetGliderStuff(playerMo)
			S_StartSound(playerMo, sfx_paglno)
		end
	else
		playerMo.letGoGliderTimer = playerMo.letGoGliderTimer or 0
		if playerMo.letGoGliderTimer > 0 then
			playerMo.letGoGliderTimer = (playerMo.letGoGliderTimer or 0) - 1
		end
	end
	
	-- print(P_IsObjectOnGround(playerMo))

	if (playerMo.paragliding or playerMo.willParaglide) and
		(
		(P_IsObjectOnGround(playerMo) or P_InQuicksand(playerMo)) or 	-- touching the ground or quicksand.
		player.playerstate == PST_DEAD						or	-- death (MORTIS)
		P_PlayerInPain(player) 								or	-- just received damage
		player.kartstuff[k_spinouttimer]  > 0				or	-- in spinout
		player.kartstuff[k_squishedtimer] > 0				or  -- squished
		playerMo.paraglideMobj.valid == false					-- paraglider got destroyed
		)
		then
		
		jglider.createDeadGlider(playerMo)
		jglider.resetGliderStuff(playerMo)
		S_StartSound(playerMo, sfx_paglno)
	end
end



-- Funny HUD + Helpers
local function countActiveSplitScreenPlayers()
	local number = 0

	for p in displayplayers.iterate do
		if p.valid then number = $ + 1 end
	end

	return number
end

local function whichSplitscreenPlayerIsThis(thisPlayer)
	local number = 1
	local howManyPlayers = countActiveSplitScreenPlayers()

	if howManyPlayers == number then return number
	elseif howManyPlayers > 1 then
		for p in displayplayers.iterate do
			if p == thisPlayer then break
			else number = $ + 1
			end
		end
	end 

	return number
end

local function isPlayerItemBoxOccupied(player)

	if 	player.kartstuff[k_itemroulette] 	> 0 or 	-- under any roulette
		player.kartstuff[k_eggmanexplode]	> 0 or	-- getting eggman'd
		player.kartstuff[k_itemtype]       ~= 0	or	-- has item (-1 is sad face)
		player.kartstuff[k_curshield]		> 0 or	-- using thunder shield
		player.kartstuff[k_stealingtimer] 	> 0 or 	-- stealing something
		player.kartstuff[k_stolentimer]		> 0 or 	-- getting stolen from
		player.kartstuff[k_growshrinktimer] > 0 or  -- beeg
		player.kartstuff[k_rocketsneakertimer] > 0 or -- rocket sneakers
		player.kartstuff[k_itemheld] == true		-- dragging/deployed item
		
		return true
	end

	return false
end

local function GliderHUDCode(v, stplyr, cam)
	if PARAGLIDERLOGICVERSION < jglider.logicVer then
		return
	end

	if stplyr 			== nil 		or
		stplyr.valid 	== false 	or 
		stplyr.mo 		== nil 		or
		stplyr.mo.valid == false 	then return end

	if stplyr.mo.paragliding == true then
		local splitscreenNum = countActiveSplitScreenPlayers()

		local barX		= 10
		local barY 		= 10
		local barWidth 	= 40

		-- we move barY down if player has item/roulette/held/growth timer
		if isPlayerItemBoxOccupied(stplyr) then
			barY = $ + 42
		end

		local flags		= V_SNAPTOTOP|V_HUDTRANS;
		local thisPlayer = whichSplitscreenPlayerIsThis(stplyr)

		if splitscreenNum > 1 then -- more than one
			if splitscreenNum == 2 then 	-- 2 players
				flags = $|V_SNAPTOLEFT 

				if thisPlayer == 2 then
					barY = $ + 108 			-- half the regular height + 8 offset pixels.
				end

			elseif splitscreenNum > 2 then	-- more than 2
				-- base pos values as ref. to the the smol item boxes
				barX 		= $ / 2
				barY 		= $ / 2
				barWidth 	= $ / 2 -- make it smol like the item box

				-- the hud snaps to the right on players 2 and 4
				-- left otherwise
				if thisPlayer % 2 == 0 then
					flags = $|V_SNAPTORIGHT
					barX  = 320 - barWidth - $
				else
					flags = $|V_SNAPTOLEFT
				end

				-- players 3 and 4 (lower ones) have their regular weird offset.
				if thisPlayer >= 3 then barY = $ + 108 end
			end
		else flags = $|V_SNAPTOLEFT
		end
		
		-- v.drawNum(barX, barY, thisPlayer, flags)
		local timer = stplyr.mo.letGoGliderTimer or 0
		if timer > 0 then
			local funnyCalc1 = FixedMul(FixedDiv(FixedDiv((barWidth * 100) * FRACUNIT, LETGOGLIDERTIMER * FRACUNIT), 100 * FRACUNIT), timer * FRACUNIT)
			local visualTimer = FixedInt(funnyCalc1)

			local barActualWidth = barWidth - visualTimer
			-- bg
			v.drawFill(barX, 	barY-1, barWidth, 		1, 21|flags)
			v.drawFill(barX-1, 	barY, 	barWidth + 2, 	1, 23|flags)
			v.drawFill(barX-1, 	barY+1, barWidth + 2, 	1, 24|flags) 
			v.drawFill(barX-1, 	barY+2, barWidth + 2, 	1, 25|flags)
			v.drawFill(barX, 	barY+3, barWidth, 		1, 26|flags)
			-- these parts move
			v.drawFill(barX+1, 	barY, 	barActualWidth - 2, 1, 7|flags)
			v.drawFill(barX, 	barY+1, barActualWidth, 	1, 9|flags)
			v.drawFill(barX+1, 	barY+2, barActualWidth - 2, 1, 12|flags)
		end
	end

	if jglider.CVar_ParagliderDebug.value == 1 then
		local consoleplayerMo = consoleplayer.mo

		v.drawString(0, 0, "Paragliding? " 		+ consoleplayerMo.paragliding)
		v.drawString(0, 10, "Will Paraglide? " 	+ consoleplayerMo.willParaglide)

		if (consoleplayerMo.paragliding) then
			v.drawString(0, 20, "momz: " + consoleplayerMo.momz)
			local oldMoMz = consoleplayerMo.oldMoMz or 0
			v.drawString(0, 30, "Old momz: " + oldMoMz)
			v.drawString(0, 40, "Vert.Mov.Inf: " + consoleplayerMo.glideVerticalMovement)
			v.drawString(0, 50, "Player Angle:\t" + consoleplayerMo.angle / ANG1)
			local turnRate = jglider.getTurnRateGlider(consoleplayer)
			local momAngle = R_PointToAngle2(0, 0, consoleplayerMo.momx, consoleplayerMo.momy)
			v.drawString(0, 60, "Momentum Angle:\t" + momAngle / ANG1)
			v.drawString(0, 70, "Angle Left:\t" + (abs(consoleplayerMo.angle) - abs(momAngle)) / ANG1)
			v.drawString(0, 80, "Turn Rate:\t" + turnRate / (ANG1 / 35))
		end
	end
end
hud.add(GliderHUDCode, "game")

if jglider.logicVer == nil then
--[[
	This looks strange, but here's the reasoning:
	As the Lua gets loaded multiple times from loaded maps, more and more
	hooks to the same stuff will happen. This is not performant.
	Checking nil here lets us know that we're the first script to load.
	And the functions attached to the hooks are on our jglider table.
	This also means we're not wasting memory defining the same functions.
	As more of the same script loads, only the function 
	changes assuming it's new, and we're not hooking more.
	It's even safer given that maps are usually loaded at the start.
]]
	addHook("MobjThinker", 		jglider.glidingLogic,		MT_PLAYER)
	addHook("LinedefExecute", 	jglider.enableParagliding, 	"PARAGLIDENABLE")
	addHook("LinedefExecute", 	jglider.removeParaglide, 	"PARAGLIDREMOVE")
end

jglider.logicVer = PARAGLIDERLOGICVERSION

else
	if jglider.CVar_ParagliderDebug ~= nil and jglider.CVar_ParagliderDebug.value == 1 then
		print("\130GLIDER WARNING:\128 An older/same paraglider LOGIC version ("..PARAGLIDERLOGICVERSION..") is being loaded, "..
				"current is "..jglider.logicVer..", skipping...")
	end
end