-- This lua file defines the paraglider object

-- this action is visual only!
-- actor.target     indicates the player it is attached to

-- if paraglide angle == player angle, idle state.
-- if not, then it is moving.
-- > 0 means it is moving to the right.
-- < 0 means it is moving to the left.
-- 1 - 5    = B frame, slight   turn
-- 5 - 10   = C frame, moderate turn
-- 10 - 15  = D frame, strong   turn
-- 15 - 20  = E frame, forceful turn

-- IMPORTANT!! NEWER VERSION WILL OVERRIDE OLDER VERSIONS
local PARAGLIDERVISUALSVERSION	= 2
if jglider == nil then
	rawset(_G, "jglider", {})
end

if jglider.visualsVer == nil or jglider.visualsVer < PARAGLIDERVISUALSVERSION then

local ANG1			        = ANG1
local ANGLE_180				= ANGLE_180
local FRACUNIT				= FRACUNIT

local paraglideSlightTurn   = ANG1
local paraglideModerateTurn = ANG1 * 15
local paraglideStrongTurn   = ANG1 * 30
local paraglideForcefulTurn = ANG1 * 45

function A_JugParaglideMovement(actor, var1, var2)

	if 	actor				== nil		or
		actor.valid			== false	or
		actor.target 		== nil 		or
		actor.target.valid 	== false 	then
		P_RemoveMobj(actor)
		return
	end

	local playerMo = actor.target

	-- position the glider above the player at all times.
	local initialPos = FixedMul(16 * FRACUNIT, playerMo.scale) * P_MobjFlip(playerMo)

	P_TeleportMove(actor, playerMo.x, playerMo.y, playerMo.z + initialPos)

	-- handle that rotation
	local isNegative = false

	local turnDifference = actor.angle - playerMo.angle
	if turnDifference < 0 then
		isNegative = true
	end

	turnDifference = abs(turnDifference)

	-- set angle to the playerMo's momentum
	local thrustAngle = R_PointToAngle2(0, 0, playerMo.momx, playerMo.momy)

	local angleCorrectRate = ANG1
	local angleMovementResult = angleCorrectRate

	if thrustAngle > playerMo.angle then
		thrustAngle = $ - angleMovementResult
	elseif thrustAngle < playerMo.angle then
		thrustAngle = $ + angleMovementResult
	end

	actor.angle = thrustAngle

	if actor.state == S_PAGL_IDLE then
		
		if turnDifference < angleCorrectRate then
			angleMovementResult = turnDifference
		end

		if turnDifference >= ANG1 then
			
			if turnDifference > paraglideForcefulTurn       then
				actor.frame = E
			elseif turnDifference > paraglideStrongTurn     then
				actor.frame = D
			elseif turnDifference > paraglideModerateTurn   then
				actor.frame = C
			elseif turnDifference > paraglideSlightTurn     then
				actor.frame = B
			end
			
			if isNegative then
				actor.frame = $ + E
			end
		else
			actor.frame = A
		end
	end

	-- handling flipping and color matching from the playerMo
	K_MatchGenericExtraFlags(actor, playerMo)

	if playerMo.player.kartstuff[k_invincibilitytimer] or
		playerMo.player.kartstuff[k_growshrinktimer] then

		actor.color = playerMo.color

		if playerMo.colorized then
			actor.colorized = true
		end

		if playerMo.player.kartstuff[k_invincibilitytimer] == 1 or
			playerMo.player.kartstuff[k_growshrinktimer] == 1 then
			actor.color = playerMo.player.skincolor
			actor.colorized = false
		end
	end

	-- handling scale
	if actor.destscale != playerMo.destscale then
		actor.destscale = playerMo.destscale
	end

	-- print(turnDifference)

end

freeslot("MT_JUG_PARAGLIDER", "MT_JUG_DEADGLIDER", "SPR_PAGL", "S_PAGL_SPAWNIDLE", "S_PAGL_SPAWN1",
"S_PAGL_SPAWN2", "S_PAGL_SPAWN3", "S_PAGL_SPAWN4", "S_PAGL_SPAWN5", "S_PAGL_SPAWN6", "S_PAGL_SPAWN7",
"S_PAGL_SPAWN8", "S_PAGL_SPAWN9", "S_PAGL_SPAWN10", "S_PAGL_IDLE", "sfx_paglgo", "sfx_paglno",
"S_PAGLRIP_DYING", "S_PAGLRIP_DEADIDLE", "S_PAGLRIP_FADE1", "S_PAGLRIP_FADE2", "S_PAGLRIP_FADE3" )

mobjinfo[MT_JUG_PARAGLIDER] =
{
	doomednum       = -1,           -- has no attached thing
	spawnstate      = S_PAGL_SPAWNIDLE,
	spawnhealth     = 1,            -- can be shot down with an item!

	radius          = 16 * FRACUNIT,
	height          = 8 * FRACUNIT,
	flags           = MF_SHOOTABLE|MF_NOGRAVITY,

	dispoffset      = 1,             -- rendered over the sprite
	-- raisestate is used to know who to attach to
}

states[S_PAGL_SPAWNIDLE] =
{
	sprite = SPR_PAGL,
	frame = J,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_SPAWNIDLE
}

states[S_PAGL_SPAWN1] =
{
	sprite = SPR_PAGL,
	frame = J,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_SPAWN2
}

states[S_PAGL_SPAWN2] =
{
	sprite = SPR_PAGL,
	frame = J,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_SPAWN3
}

states[S_PAGL_SPAWN3] =
{
	sprite = SPR_PAGL,
	frame = K,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_SPAWN4
}

states[S_PAGL_SPAWN4] =
{
	sprite = SPR_PAGL,
	frame = K,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_SPAWN5
}

states[S_PAGL_SPAWN5] =
{
	sprite = SPR_PAGL,
	frame = L,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_SPAWN6
}

states[S_PAGL_SPAWN6] =
{
	sprite = SPR_PAGL,
	frame = L,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_SPAWN7
}

states[S_PAGL_SPAWN7] =
{
	sprite = SPR_PAGL,
	frame = M,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_SPAWN8
}

states[S_PAGL_SPAWN8] =
{
	sprite = SPR_PAGL,
	frame = M,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_SPAWN9
}

states[S_PAGL_SPAWN9] =
{
	sprite = SPR_PAGL,
	frame = N,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_SPAWN10
}

states[S_PAGL_SPAWN10] =
{
	sprite = SPR_PAGL,
	frame = N,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_IDLE
}

states[S_PAGL_IDLE] =
{
	sprite = SPR_PAGL,
	frame = A,
	tics = 1,
	action = A_JugParaglideMovement,
	nextstate = S_PAGL_IDLE
}


mobjinfo[MT_JUG_DEADGLIDER] =
{
	doomednum       = -1,           -- has no attached thing
	spawnstate      = S_PAGLRIP_DYING,
	spawnhealth     = 1,

	radius          = 8 * FRACUNIT,
	height          = 4 * FRACUNIT,
	flags           = MF_NOCLIPTHING,

	dispoffset      = 1,             -- rendered over the sprite
}

states[S_PAGLRIP_DYING] =
{
	sprite = SPR_PAGL,
	frame = FF_ANIMATE|O,
	tics = 12,
	var1 = 5,
	var2 = 2,
	nextstate = S_PAGLRIP_DEADIDLE
}

states[S_PAGLRIP_DEADIDLE] =
{
	sprite = SPR_PAGL,
	frame = T,
	tics = 140,
	nextstate = S_PAGLRIP_FADE1
}

states[S_PAGLRIP_FADE1] =
{
	sprite = SPR_PAGL,
	frame = TR_TRANS30|T,
	tics = 2,
	nextstate = S_PAGLRIP_FADE2
}

states[S_PAGLRIP_FADE2] =
{
	sprite = SPR_PAGL,
	frame = TR_TRANS60|T,
	tics = 2,
	nextstate = S_PAGLRIP_FADE3
}

states[S_PAGLRIP_FADE3] =
{
	sprite = SPR_PAGL,
	frame = TR_TRANS90|T,
	tics = 2,
	nextstate = S_NULL
}

sfxinfo[sfx_paglgo] = {false, 64, SF_TOTALLYSINGLE}
sfxinfo[sfx_paglno] = {false, 64, SF_TOTALLYSINGLE}

jglider.gliderInstaShieldLogic = function(target, inflictor, source, damagetype)
	-- we generally don't tend to care how we got damaged
	-- we just need to check if it applies when the player is
	-- invulnerable, unless it's nil (caused by the world)

	if inflictor == nil and source == nil then return true end

	local player = target.target
	if (player.powers[pw_flashing] or player.kartstuff[k_invincibilitytimer]) then
		K_DoInstashield(player)
		return false
	end

end

if jglider.visualsVer == nil then
	addHook("MobjDamage", jglider.gliderInstaShieldLogic, MT_JUG_PARAGLIDER)
end

-- let's update the version
jglider.visualsVer = PARAGLIDERVISUALSVERSION
else
	if jglider.CVar_ParagliderDebug ~= nil and jglider.CVar_ParagliderDebug.value == 1 then
		print("\130GLIDER WARNING:\128 An older/same paraglider VISUALS version ("..PARAGLIDERVISUALSVERSION..") is being loaded, "..
				"current is "..jglider.visualsVer..", skipping...")
	end
end